from __future__ import annotations

from typing import List, Optional, Any, Literal, overload

from ai21.clients.studio.resources.studio_resource import StudioResource
from ai21.clients.studio.resources.chat.base_chat_completions import BaseChatCompletions
from ai21.models import ChatMessage as J2ChatMessage
from ai21.models.chat import ChatCompletionResponse, ChatCompletionChunk
from ai21.models.chat.chat_message import ChatMessageParam
from ai21.models.chat.document_schema import DocumentSchema
from ai21.models.chat.response_format import ResponseFormat
from ai21.models.chat.tool_defintions import ToolDefinition
from ai21.stream.stream import Stream
from ai21.types import NotGiven, NOT_GIVEN

__all__ = ["ChatCompletions"]


class ChatCompletions(StudioResource, BaseChatCompletions):
    @overload
    def create(
        self,
        model: str,
        messages: List[ChatMessageParam],
        max_tokens: int | NotGiven = NOT_GIVEN,
        temperature: float | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
        stop: str | List[str] | NotGiven = NOT_GIVEN,
        n: int | NotGiven = NOT_GIVEN,
        stream: Optional[Literal[False]] | NotGiven = NOT_GIVEN,
        tools: List[ToolDefinition] | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        documents: List[DocumentSchema] | NotGiven = NOT_GIVEN,
        **kwargs: Any,
    ) -> ChatCompletionResponse:
        pass

    @overload
    def create(
        self,
        model: str,
        messages: List[ChatMessageParam],
        stream: Literal[True],
        max_tokens: int | NotGiven = NOT_GIVEN,
        temperature: float | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
        stop: str | List[str] | NotGiven = NOT_GIVEN,
        n: int | NotGiven = NOT_GIVEN,
        tools: List[ToolDefinition] | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        documents: List[DocumentSchema] | NotGiven = NOT_GIVEN,
        **kwargs: Any,
    ) -> Stream[ChatCompletionChunk]:
        pass

    def create(
        self,
        messages: List[ChatMessageParam],
        model: Optional[str] = None,
        max_tokens: int | NotGiven = NOT_GIVEN,
        temperature: float | NotGiven = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
        stop: str | List[str] | NotGiven = NOT_GIVEN,
        n: int | NotGiven = NOT_GIVEN,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        tools: List[ToolDefinition] | NotGiven = NOT_GIVEN,
        response_format: ResponseFormat | NotGiven = NOT_GIVEN,
        documents: List[DocumentSchema] | NotGiven = NOT_GIVEN,
        **kwargs: Any,
    ) -> ChatCompletionResponse | Stream[ChatCompletionChunk]:
        if any(isinstance(item, J2ChatMessage) for item in messages):
            raise ValueError(
                "Please use the ChatMessage class from ai21.models.chat"
                " instead of ai21.models when working with chat completions."
            )

        body = self._create_body(
            model=self._get_model(model=model, model_id=kwargs.pop("model_id", None)),
            messages=messages,
            stop=stop,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            n=n,
            stream=stream or False,
            tools=tools,
            response_format=response_format,
            documents=documents,
            **kwargs,
        )

        return self._post(
            path=f"/{self._module_name}",
            body=body,
            stream=stream or False,
            stream_cls=Stream[ChatCompletionChunk],
            response_cls=ChatCompletionResponse,
        )



from typing import Any, Dict, get_args, cast

from ai21.types import NotGiven


def is_not_given(value: Any) -> bool:
    return isinstance(value, NotGiven)


def remove_not_given(body: Dict[str, Any]) -> Dict[str, Any]:
    return {k: v for k, v in body.items() if not is_not_given(v)}


def to_camel_case(snake_str: str) -> str:
    return "".join(x.capitalize() for x in snake_str.lower().split("_"))


def to_lower_camel_case(snake_str: str) -> str:
    # We capitalize the first letter of each component except the first one
    # with the 'capitalize' method and join them together.
    camel_string = to_camel_case(snake_str)
    return snake_str[0].lower() + camel_string[1:]


def extract_type(type_to_extract: Any) -> type:
    args = get_args(type_to_extract)
    try:
        return cast(type, args[0])
    except IndexError as err:
        raise RuntimeError(
            f"Expected type {type_to_extract} to have a type argument at index 0 but it did not"
        ) from err



from __future__ import annotations

import warnings
from abc import ABC
from typing import List, Optional, Union, Any, Dict, Literal

from ai21.models.chat.chat_message import ChatMessageParam
from ai21.models.chat.document_schema import DocumentSchema
from ai21.models.chat.response_format import ResponseFormat
from ai21.models.chat.tool_defintions import ToolDefinition
from ai21.types import NotGiven
from ai21.utils.typing import remove_not_given
from ai21.models._pydantic_compatibility import _to_dict


class BaseChatCompletions(ABC):
    _module_name = "chat/completions"

    def _get_model(self, model: Optional[str], model_id: Optional[str]) -> str:
        if model_id is not None:
            warnings.warn(
                "The 'model_id' parameter is deprecated and will be removed in a future version."
                " Please use 'model' instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        if model_id and model:
            raise ValueError("Please provide only 'model' as 'model_id' is deprecated.")

        if not model and not model_id:
            raise ValueError("model should be provided 'create' method call")

        return model or model_id

    def _create_body(
        self,
        model: str,
        messages: List[ChatMessageParam],
        max_tokens: Optional[int] | NotGiven,
        temperature: Optional[float] | NotGiven,
        top_p: Optional[float] | NotGiven,
        stop: Optional[Union[str, List[str]]] | NotGiven,
        n: Optional[int] | NotGiven,
        stream: Literal[False] | Literal[True] | NotGiven,
        tools: List[ToolDefinition] | NotGiven,
        response_format: ResponseFormat | NotGiven,
        documents: List[DocumentSchema] | NotGiven,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        return remove_not_given(
            {
                "model": model,
                "messages": [_to_dict(message) for message in messages],
                "temperature": temperature,
                "max_tokens": max_tokens,
                "top_p": top_p,
                "stop": stop,
                "n": n,
                "stream": stream,
                "tools": tools,
                "response_format": response_format,
                "documents": documents,
                **kwargs,
            }
        )



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import List, Dict, Optional

from ai21.models import Penalty, CompletionsResponse
from ai21.types import NOT_GIVEN, NotGiven
from ai21.utils.typing import remove_not_given
from ai21.models._pydantic_compatibility import _to_dict


class Completion(ABC):
    _module_name = "complete"

    def _get_model(self, model: Optional[str], model_id: Optional[str]) -> str:
        if model_id is not None:
            warnings.warn(
                "The 'model_id' parameter is deprecated and will be removed in a future version."
                " Please use 'model' instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        if model_id and model:
            raise ValueError("Please provide only 'model' as 'model_id' is deprecated.")

        if not model and not model_id:
            raise ValueError("model should be provided 'create' method call")

        return model or model_id

    @abstractmethod
    def create(
        self,
        model: str,
        prompt: str,
        *,
        max_tokens: int | NotGiven = NOT_GIVEN,
        num_results: int | NotGiven = NOT_GIVEN,
        min_tokens: int | NotGiven = NOT_GIVEN,
        temperature: float | NOT_GIVEN = NOT_GIVEN,
        top_p: float | NotGiven = NOT_GIVEN,
        top_k_return: int | NotGiven = NOT_GIVEN,
        stop_sequences: List[str] | NotGiven = NOT_GIVEN,
        frequency_penalty: Penalty | NotGiven = NOT_GIVEN,
        presence_penalty: Penalty | NotGiven = NOT_GIVEN,
        count_penalty: Penalty | NotGiven = NOT_GIVEN,
        epoch: int | NotGiven = NOT_GIVEN,
        logit_bias: Dict[str, float] | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> CompletionsResponse:
        """
        :param model: model type you wish to interact with
        :param prompt: Text for model to complete
        :param max_tokens: The maximum number of tokens to generate per result
        :param num_results: Number of completions to sample and return.
        :param min_tokens: The minimum number of tokens to generate per result.
        :param temperature: A value controlling the "creativity" of the model's responses.
        :param top_p: A value controlling the diversity of the model's responses.
        :param top_k_return: The number of top-scoring tokens to consider for each generation step.
        :param stop_sequences: Stops decoding if any of the strings is generated
        :param frequency_penalty: A penalty applied to tokens that are frequently generated.
        :param presence_penalty:  A penalty applied to tokens that are already present in the prompt.
        :param count_penalty: A penalty applied to tokens based on their frequency in the generated responses
        :param epoch:
        :param logit_bias: A dictionary which contains mapping from strings to floats, where the strings are text
        representations of the tokens and the floats are the biases themselves. A positive bias increases generation
        probability for a given token and a negative bias decreases it.
        :param kwargs:
        :return:
        """
        pass

    def _create_body(
        self,
        model: str,
        prompt: str,
        max_tokens: int | NotGiven,
        num_results: int | NotGiven,
        min_tokens: int | NotGiven,
        temperature: float | NotGiven,
        top_p: float | NotGiven,
        top_k_return: int | NotGiven,
        stop_sequences: List[str] | NotGiven,
        frequency_penalty: Penalty | NotGiven,
        presence_penalty: Penalty | NotGiven,
        count_penalty: Penalty | NotGiven,
        epoch: int | NotGiven,
        logit_bias: Dict[str, float] | NotGiven,
        **kwargs,
    ):
        return remove_not_given(
            {
                "model": model,
                "prompt": prompt,
                "maxTokens": max_tokens,
                "numResults": num_results,
                "minTokens": min_tokens,
                "temperature": temperature,
                "topP": top_p,
                "topKReturn": top_k_return,
                "stopSequences": stop_sequences,
                "frequencyPenalty": (NOT_GIVEN if frequency_penalty is NOT_GIVEN else _to_dict(frequency_penalty)),
                "presencePenalty": (NOT_GIVEN if presence_penalty is NOT_GIVEN else _to_dict(presence_penalty)),
                "countPenalty": (NOT_GIVEN if count_penalty is NOT_GIVEN else _to_dict(count_penalty)),
                "epoch": epoch,
                "logitBias": logit_bias,
                **kwargs,
            }
        )

    def _get_completion_path(self, model: str):
        return f"/{model}/{self._module_name}"



from __future__ import annotations
from typing import List

from ai21.clients.common.conversational_rag import ConversationalRag
from ai21.clients.studio.resources.studio_resource import (
    AsyncStudioResource,
    StudioResource,
)
from ai21.models.chat import ChatMessage
from ai21.models.responses.conversational_rag_response import ConversationalRagResponse
from ai21.models.retrieval_strategy import RetrievalStrategy
from ai21.types import NotGiven, NOT_GIVEN


class StudioConversationalRag(StudioResource, ConversationalRag):
    def create(
        self,
        messages: List[ChatMessage],
        *,
        path: str | NotGiven = NOT_GIVEN,
        labels: List[str] | NotGiven = NOT_GIVEN,
        file_ids: List[str] | NotGiven = NOT_GIVEN,
        max_segments: int | NotGiven = NOT_GIVEN,
        retrieval_strategy: RetrievalStrategy | str | NotGiven = NOT_GIVEN,
        retrieval_similarity_threshold: float | NotGiven = NOT_GIVEN,
        max_neighbors: int | NotGiven = NOT_GIVEN,
        hybrid_search_alpha: float | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> ConversationalRagResponse:
        body = self._create_body(
            messages=messages,
            path=path,
            labels=labels,
            file_ids=file_ids,
            max_segments=max_segments,
            retrieval_strategy=retrieval_strategy,
            retrieval_similarity_threshold=retrieval_similarity_threshold,
            max_neighbors=max_neighbors,
            hybrid_search_alpha=hybrid_search_alpha,
            **kwargs,
        )

        return self._post(path=f"/{self._module_name}", body=body, response_cls=ConversationalRagResponse)


class AsyncStudioConversationalRag(AsyncStudioResource, ConversationalRag):
    async def create(
        self,
        messages: List[ChatMessage],
        *,
        path: str | NotGiven = NOT_GIVEN,
        labels: List[str] | NotGiven = NOT_GIVEN,
        file_ids: List[str] | NotGiven = NOT_GIVEN,
        max_segments: int | NotGiven = NOT_GIVEN,
        retrieval_strategy: RetrievalStrategy | str | NotGiven = NOT_GIVEN,
        retrieval_similarity_threshold: float | NotGiven = NOT_GIVEN,
        max_neighbors: int | NotGiven = NOT_GIVEN,
        hybrid_search_alpha: float | NotGiven = NOT_GIVEN,
        **kwargs,
    ) -> ConversationalRagResponse:
        body = self._create_body(
            messages=messages,
            path=path,
            labels=labels,
            file_ids=file_ids,
            max_segments=max_segments,
            retrieval_strategy=retrieval_strategy,
            retrieval_similarity_threshold=retrieval_similarity_threshold,
            max_neighbors=max_neighbors,
            hybrid_search_alpha=hybrid_search_alpha,
            **kwargs,
        )

        return await self._post(path=f"/{self._module_name}", body=body, response_cls=ConversationalRagResponse)



from datetime import date
from typing import Optional, List

from pydantic import Field

from ai21.models.ai21_base_model import AI21BaseModel


class FileResponse(AI21BaseModel):
    file_id: str = Field(alias="fileId")
    name: str
    file_type: str = Field(alias="fileType")
    size_bytes: int = Field(alias="sizeBytes")
    created_by: str = Field(alias="createdBy")
    creation_date: date = Field(alias="creationDate")
    last_updated: date = Field(alias="lastUpdated")
    status: str
    path: Optional[str] = None
    labels: Optional[List[str]] = None
    public_url: Optional[str] = Field(default=None, alias="publicUrl")
    error_code: Optional[int] = Field(default=None, alias="errorCode")
    error_message: Optional[str] = Field(default=None, alias="errorMessage")



class BedrockModelID:
    J2_MID_V1 = "ai21.j2-mid-v1"
    J2_ULTRA_V1 = "ai21.j2-ultra-v1"
    JAMBA_INSTRUCT_V1 = "ai21.jamba-instruct-v1:0"
    JAMBA_1_5_MINI = "ai21.jamba-1-5-mini-v1:0"
    JAMBA_1_5_LARGE = "ai21.jamba-1-5-large-v1:0"



import logging
import os
import re

from ai21.ai21_env_config import AI21EnvConfig


_verbose = False

logger = logging.getLogger("ai21")
httpx_logger = logging.getLogger("httpx")


class CensorSecretsFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        # Get original log message
        message = super().format(record)

        if not get_verbose():
            return self._censor_secrets(message)

        return message

    def _censor_secrets(self, message: str) -> str:
        # Regular expression to find the Authorization key and its value
        pattern = r"('Authorization':\s*'[^']*'|'api-key':\s*'[^']*'|'X-Amz-Security-Token':\s*'[^']*')"

        def replacement(match):
            return match.group(0).split(":")[0] + ": '**************'"

        # Substitute the Authorization value with **************
        return re.sub(pattern, replacement, message)


def set_verbose(value: bool) -> None:
    """
    Use this function if you want to log additional, more sensitive data like - secrets and environment variables.
    Log level will be set to DEBUG if verbose is set to True.
    """
    global _verbose
    _verbose = value

    set_debug(_verbose)

    AI21EnvConfig.log(with_secrets=value)


def set_debug(value: bool) -> None:
    """
    Additional way to set log level to DEBUG.
    """
    if value:
        os.environ["AI21_LOG_LEVEL"] = "debug"
    else:
        os.environ["AI21_LOG_LEVEL"] = "info"

    setup_logger()


def get_verbose() -> bool:
    global _verbose
    return _verbose


def setup_logger() -> None:
    handler = logging.StreamHandler()

    handler.setFormatter(
        CensorSecretsFormatter(fmt="[%(asctime)s - %(name)s - %(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    )

    logger.addHandler(handler)
    httpx_logger.addHandler(handler)

    if AI21EnvConfig.log_level.lower() == "debug":
        logger.setLevel(logging.DEBUG)
        httpx_logger.setLevel(logging.DEBUG)
    elif AI21EnvConfig.log_level.lower() == "info":
        logger.setLevel(logging.INFO)



from typing import Dict

from ai21_tokenizer import Tokenizer, PreTrainedTokenizers

from ai21.tokenizers.ai21_tokenizer import AI21Tokenizer, AsyncAI21Tokenizer

_cached_tokenizers: Dict[str, AI21Tokenizer] = {}
_cached_async_tokenizers: Dict[str, AsyncAI21Tokenizer] = {}


def get_tokenizer(name: str = PreTrainedTokenizers.J2_TOKENIZER) -> AI21Tokenizer:
    """
    Get the tokenizer instance.

    If the tokenizer instance is not cached, it will be created using the Tokenizer.get_tokenizer() method.
    """
    global _cached_tokenizers

    if _cached_tokenizers.get(name) is None:
        _cached_tokenizers[name] = AI21Tokenizer(Tokenizer.get_tokenizer(name))

    return _cached_tokenizers[name]


async def get_async_tokenizer(name: str = PreTrainedTokenizers.J2_TOKENIZER) -> AsyncAI21Tokenizer:
    """
    Get the async tokenizer instance.

    If the tokenizer instance is not cached, it will be created using the Tokenizer.get_tokenizer() method.
    """
    global _cached_async_tokenizers

    if _cached_async_tokenizers.get(name) is None:
        _cached_async_tokenizers[name] = AsyncAI21Tokenizer(await Tokenizer.get_async_tokenizer(name))

    return _cached_async_tokenizers[name]



from typing import Any

from ai21.ai21_env_config import AI21EnvConfig
from ai21.clients.azure.ai21_azure_client import AI21AzureClient, AsyncAI21AzureClient
from ai21.clients.studio.ai21_client import AI21Client
from ai21.clients.studio.async_ai21_client import AsyncAI21Client

from ai21.errors import (
    AI21APIError,
    APITimeoutError,
    MissingApiKeyError,
    ModelPackageDoesntExistError,
    AI21Error,
    TooManyRequestsError,
)
from ai21.logger import setup_logger
from ai21.version import VERSION

__version__ = VERSION
setup_logger()


def _import_bedrock_client():
    from ai21.clients.bedrock.ai21_bedrock_client import AI21BedrockClient

    return AI21BedrockClient


def _import_bedrock_model_id():
    from ai21.clients.bedrock.bedrock_model_id import BedrockModelID

    return BedrockModelID


def _import_async_bedrock_client():
    from ai21.clients.bedrock.ai21_bedrock_client import AsyncAI21BedrockClient

    return AsyncAI21BedrockClient


def _import_vertex_client():
    from ai21.clients.vertex.ai21_vertex_client import AI21VertexClient

    return AI21VertexClient


def _import_async_vertex_client():
    from ai21.clients.vertex.ai21_vertex_client import AsyncAI21VertexClient

    return AsyncAI21VertexClient


def __getattr__(name: str) -> Any:
    try:
        if name == "AI21BedrockClient":
            return _import_bedrock_client()

        if name == "BedrockModelID":
            return _import_bedrock_model_id()

        if name == "AsyncAI21BedrockClient":
            return _import_async_bedrock_client()

        if name == "AI21VertexClient":
            return _import_vertex_client()

        if name == "AsyncAI21VertexClient":
            return _import_async_vertex_client()
    except ImportError as e:
        raise ImportError('Please install "ai21[AWS]" for Bedrock, or "ai21[Vertex]" for Vertex') from e


__all__ = [
    "AI21EnvConfig",
    "AI21Client",
    "AsyncAI21Client",
    "AI21APIError",
    "APITimeoutError",
    "AI21Error",
    "MissingApiKeyError",
    "ModelPackageDoesntExistError",
    "TooManyRequestsError",
    "AI21BedrockClient",
    "BedrockModelID",
    "AI21AzureClient",
    "AsyncAI21AzureClient",
    "AsyncAI21BedrockClient",
    "AI21VertexClient",
    "AsyncAI21VertexClient",
]



from __future__ import annotations

import json
from abc import ABC
from typing import Any, Dict, Optional, BinaryIO, get_origin

import httpx

from ai21.http_client.async_http_client import AsyncAI21HTTPClient
from ai21.http_client.http_client import AI21HTTPClient
from ai21.models._pydantic_compatibility import _from_dict
from ai21.stream.stream_commons import _SSEDecoderBase
from ai21.types import ResponseT, StreamT, AsyncStreamT
from ai21.utils.typing import extract_type


def _cast_response(
    response: httpx.Response,
    response_cls: Optional[ResponseT],
    stream_cls: Optional[AsyncStreamT] = None,
    stream: bool = False,
    streaming_decoder: Optional[_SSEDecoderBase] = None,
) -> ResponseT | AsyncStreamT | None:
    if stream and stream_cls is not None:
        cast_to = extract_type(stream_cls)
        return stream_cls(cast_to=cast_to, response=response, streaming_decoder=streaming_decoder)

    if response_cls is None:
        return None

    if response_cls == dict:
        return response.json()

    if response_cls == str:
        return json.loads(response.json())

    origin_type = get_origin(response_cls)

    if origin_type is not None and origin_type == list:
        subtype = extract_type(response_cls)
        return [_from_dict(obj=subtype, obj_dict=item) for item in response.json()]

    return _from_dict(obj=response_cls, obj_dict=response.json())


class StudioResource(ABC):
    def __init__(self, client: AI21HTTPClient):
        self._client = client

    def _post(
        self,
        path: str,
        body: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        response_cls: Optional[ResponseT] = None,
        stream_cls: Optional[StreamT] = None,
        stream: bool = False,
        files: Optional[Dict[str, BinaryIO]] = None,
    ) -> ResponseT | StreamT:
        response = self._client.execute_http_request(
            method="POST",
            path=path,
            stream=stream,
            body=body or {},
            params=params or {},
            files=files,
        )

        return _cast_response(
            stream=stream,
            response=response,
            response_cls=response_cls,
            stream_cls=stream_cls,
            streaming_decoder=self._client._get_streaming_decoder(),
        )

    def _get(
        self, path: str, response_cls: Optional[ResponseT] = None, params: Optional[Dict[str, Any]] = None
    ) -> ResponseT | StreamT:
        response = self._client.execute_http_request(method="GET", path=path, params=params or {})
        return _cast_response(response=response, response_cls=response_cls)

    def _put(
        self, path: str, response_cls: Optional[ResponseT] = None, body: Dict[str, Any] = None
    ) -> ResponseT | StreamT:
        response = self._client.execute_http_request(method="PUT", path=path, body=body or {})
        return _cast_response(response=response, response_cls=response_cls)

    def _delete(self, path: str, response_cls: Optional[ResponseT] = None) -> ResponseT | StreamT:
        response = self._client.execute_http_request(
            method="DELETE",
            path=path,
        )
        return _cast_response(response=response, response_cls=response_cls)


class AsyncStudioResource(ABC):
    def __init__(self, client: AsyncAI21HTTPClient):
        self._client = client

    async def _post(
        self,
        path: str,
        body: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        response_cls: Optional[ResponseT] = None,
        stream_cls: Optional[AsyncStreamT] = None,
        stream: bool = False,
        files: Optional[Dict[str, BinaryIO]] = None,
    ) -> ResponseT | AsyncStreamT:
        response = await self._client.execute_http_request(
            method="POST",
            path=path,
            stream=stream,
            body=body or {},
            params=params or {},
            files=files,
        )

        return _cast_response(
            stream=stream,
            response=response,
            response_cls=response_cls,
            stream_cls=stream_cls,
            streaming_decoder=self._client._get_streaming_decoder(),
        )

    async def _get(
        self, path: str, response_cls: Optional[ResponseT] = None, params: Optional[Dict[str, Any]] = None
    ) -> ResponseT | AsyncStreamT:
        response = await self._client.execute_http_request(method="GET", path=path, params=params or {})
        return _cast_response(response=response, response_cls=response_cls)

    async def _put(
        self, path: str, response_cls: Optional[ResponseT] = None, body: Dict[str, Any] = None
    ) -> ResponseT | AsyncStreamT:
        response = await self._client.execute_http_request(method="PUT", path=path, body=body or {})
        return _cast_response(response=response, response_cls=response_cls)

    async def _delete(self, path: str, response_cls: Optional[ResponseT] = None) -> ResponseT | AsyncStreamT:
        response = await self._client.execute_http_request(
            method="DELETE",
            path=path,
        )
        return _cast_response(response=response, response_cls=response_cls)



from ai21.clients.studio.resources.studio_conversational_rag import StudioConversationalRag
from ai21.clients.studio.resources.studio_resource import StudioResource
from ai21.http_client.http_client import AI21HTTPClient


class Beta(StudioResource):
    def __init__(self, client: AI21HTTPClient):
        super().__init__(client)

        self.conversational_rag = StudioConversationalRag(client)



from typing_extensions import Literal, TypedDict, Required

from ai21.models.chat import FunctionToolDefinition


class ToolDefinition(TypedDict, total=False):
    type: Required[Literal["function"]]
    function: Required[FunctionToolDefinition]



from typing import Optional, List

from ai21.models.ai21_base_model import AI21BaseModel
from ai21.models.logprobs import Logprobs
from ai21.models.usage_info import UsageInfo


class ChoiceDelta(AI21BaseModel):
    content: Optional[str] = None
    role: Optional[str] = None


class ChoicesChunk(AI21BaseModel):
    index: int
    delta: ChoiceDelta
    logprobs: Optional[Logprobs] = None
    finish_reason: Optional[str] = None


class ChatCompletionChunk(AI21BaseModel):
    id: str
    choices: List[ChoicesChunk]
    usage: Optional[UsageInfo] = None



from typing import List, Any

from ai21_tokenizer import BaseTokenizer, AsyncBaseTokenizer


class AI21Tokenizer:
    """
    A class that wraps a tokenizer and provides additional functionality.
    """

    def __init__(self, tokenizer: BaseTokenizer):
        self._tokenizer = tokenizer

    def count_tokens(self, text: str) -> int:
        encoded_text = self._tokenizer.encode(text)

        return len(encoded_text)

    def tokenize(self, text: str, **kwargs: Any) -> List[str]:
        encoded_text = self._tokenizer.encode(text, **kwargs)

        return self._tokenizer.convert_ids_to_tokens(encoded_text, **kwargs)

    def detokenize(self, tokens: List[str], **kwargs: Any) -> str:
        token_ids = self._tokenizer.convert_tokens_to_ids(tokens)

        return self._tokenizer.decode(token_ids, **kwargs)


class AsyncAI21Tokenizer:
    """
    A class that wraps an async tokenizer and provides additional functionality.
    """

    def __init__(self, tokenizer: AsyncBaseTokenizer):
        self._tokenizer = tokenizer

    async def count_tokens(self, text: str) -> int:
        encoded_text = await self._tokenizer.encode(text)

        return len(encoded_text)

    async def tokenize(self, text: str, **kwargs: Any) -> List[str]:
        encoded_text = await self._tokenizer.encode(text, **kwargs)

        return await self._tokenizer.convert_ids_to_tokens(encoded_text, **kwargs)

    async def detokenize(self, tokens: List[str], **kwargs: Any) -> str:
        token_ids = await self._tokenizer.convert_tokens_to_ids(tokens)

        return await self._tokenizer.decode(token_ids, **kwargs)



from __future__ import annotations

import json
import platform
from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Union, Any, Optional, Dict, BinaryIO

import httpx

from ai21.errors import (
    BadRequest,
    Unauthorized,
    UnprocessableEntity,
    TooManyRequestsError,
    AI21ServerError,
    ServiceUnavailable,
    AI21APIError,
    MissingApiKeyError,
)
from ai21.models.request_options import RequestOptions
from ai21.stream.async_stream import AsyncStream
from ai21.stream.stream import Stream
from ai21.version import VERSION

_HttpxClientT = TypeVar("_HttpxClientT", bound=Union[httpx.Client, httpx.AsyncClient])
_DefaultStreamT = TypeVar("_DefaultStreamT", bound=Union[Stream[Any], AsyncStream[Any]])

DEFAULT_TIMEOUT_SEC = 300
DEFAULT_NUM_RETRIES = 0
RETRY_BACK_OFF_FACTOR = 0.5
TIME_BETWEEN_RETRIES = 1
RETRY_ERROR_CODES = (408, 429, 500, 503)
RETRY_METHOD_WHITELIST = ["GET", "POST", "PUT"]


def handle_non_success_response(status_code: int, response_text: str):
    if status_code == 400:
        raise BadRequest(details=response_text)
    if status_code == 401:
        raise Unauthorized(details=response_text)
    if status_code == 422:
        raise UnprocessableEntity(details=response_text)
    if status_code == 429:
        raise TooManyRequestsError(details=response_text)
    if status_code == 500:
        raise AI21ServerError(details=response_text)
    if status_code == 503:
        raise ServiceUnavailable(details=response_text)
    raise AI21APIError(status_code, details=response_text)


class BaseHttpClient(ABC, Generic[_HttpxClientT, _DefaultStreamT]):
    _client: Optional[_HttpxClientT] = None
    _request: Any

    def __init__(
        self,
        base_url: str,
        api_key: Optional[str] = None,
        requires_api_key: bool = True,
        timeout_sec: int = None,
        num_retries: int = None,
        headers: Dict = None,
        via: Optional[str] = None,
    ):
        if requires_api_key and api_key is None:
            raise MissingApiKeyError()

        self._base_url = base_url
        self._api_key = api_key
        self._timeout_sec = timeout_sec or DEFAULT_TIMEOUT_SEC
        self._num_retries = num_retries or DEFAULT_NUM_RETRIES
        self._headers = headers or {}
        self._apply_retry_policy = self._num_retries > 0
        self._via = via

    def _should_retry(self, response: httpx.Response) -> bool:
        return response.status_code in RETRY_ERROR_CODES and response.request.method in RETRY_METHOD_WHITELIST

    def _get_request_data(
        self,
        files: Optional[Dict[str, BinaryIO]],
        method: str,
        body=Optional[Dict],
    ) -> Dict[str, Any] | bytes:
        if files is not None:
            if method != "POST":
                raise ValueError(
                    f"execute_http_request supports only POST for files upload, but {method} was supplied instead"
                )

            return body
        else:
            return json.dumps(body).encode() if body else None

    def _get_request_headers(
        self, files: Optional[Dict[str, BinaryIO]], headers: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        if files is not None and "Content-Type" in self._headers:
            return {key: value for key, value in self._headers.items() if key != "Content-Type"}
        if headers is not None:
            return {**self._headers, **headers}
        return self._headers

    def add_headers(self, headers: Dict[str, Any]) -> None:
        self._headers.update(headers)

    @abstractmethod
    def execute_http_request(
        self,
        method: str,
        path: str,
        params: Optional[Dict] = None,
        stream: bool = False,
        files: Optional[Dict[str, BinaryIO]] = None,
        extra_headers: Optional[Dict] = None,
    ) -> httpx.Response:
        pass

    @abstractmethod
    def _run_request(
        self,
        options: RequestOptions,
    ) -> httpx.Response:
        pass

    @abstractmethod
    def _init_client(self, client: Optional[_HttpxClientT]) -> _HttpxClientT:
        pass

    def _build_headers(self, passed_headers: Optional[Dict[str, Any]]) -> Dict[str, Any]:
        headers = {
            "Content-Type": "application/json",
            "User-Agent": self._build_user_agent(),
        }

        if self._api_key:
            headers["Authorization"] = f"Bearer {self._api_key}"

        if passed_headers is not None:
            headers.update(passed_headers)

        return headers

    def _build_user_agent(self) -> str:
        user_agent = (
            f"AI21 studio SDK {VERSION} Python {platform.python_version()} Operating System {platform.platform()}"
        )

        if self._via is not None:
            user_agent = f"{user_agent} via: {self._via}"

        return user_agent

    def _build_request(self, options: RequestOptions) -> httpx.Request:
        data = self._get_request_data(files=options.files, method=options.method, body=options.body)
        headers = self._get_request_headers(files=options.files, headers=options.headers)

        return self._client.build_request(
            method=options.method,
            url=self._prepare_url(options),
            headers=headers,
            timeout=options.timeout,
            params=options.params,
            data=data,
            files=options.files,
        )

    def _prepare_url(self, options: RequestOptions) -> str:
        if options.path:
            return f"{options.url}{options.path}"

        return options.url

    def _extract_streaming_error_details(self, response: httpx.Response) -> str:
        try:
            return response.read().decode("utf-8")
        except Exception:
            return "could not extract streaming error details"



from __future__ import annotations
from typing import Optional, List

from ai21.models.chat import ChatMessage
from ai21.models.ai21_base_model import AI21BaseModel


class ConversationalRagSource(AI21BaseModel):
    text: str
    file_id: str
    file_name: str
    score: float
    order: Optional[int] = None
    public_url: Optional[str] = None
    labels: Optional[List[str]] = None


class ConversationalRagResponse(AI21BaseModel):
    id: str
    choices: List[ChatMessage]
    search_queries: Optional[List[str]]
    context_retrieved: bool
    answer_in_context: bool
    sources: List[ConversationalRagSource]



import warnings
from functools import wraps

V3_DEPRECATION_MESSAGE = "This model is not supported anymore - Please upgrade to v3.0.0"


def deprecated(message):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(message, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return decorator



from __future__ import annotations

from types import TracebackType
from typing import Generic, Iterator, Optional, Any

import httpx
from typing_extensions import Self

from ai21.stream.stream_commons import _T, _SSEDecoder, _SSE_DONE_MSG, get_stream_message


class Stream(Generic[_T]):
    response: httpx.Response

    def __init__(
        self,
        *,
        cast_to: type[_T],
        response: httpx.Response,
        streaming_decoder: Optional[Any] = None,
    ):
        self.response = response
        self.cast_to = cast_to
        self._decoder = streaming_decoder or _SSEDecoder()
        self._iterator = self.__stream__()

    def __next__(self) -> _T:
        return self._iterator.__next__()

    def __iter__(self) -> Iterator[_T]:
        for item in self._iterator:
            yield item

    def __stream__(self) -> Iterator[_T]:
        iterator = self._decoder.iter(self.response)
        for chunk in iterator:
            if chunk.endswith(_SSE_DONE_MSG):
                break

            yield get_stream_message(chunk, self.cast_to)

        # Ensure the entire stream is consumed
        for _chunk in iterator:
            ...

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.close()

    def close(self):
        self.response.close()



from __future__ import annotations

from typing import Optional

from ai21.models.ai21_base_model import AI21BaseModel
from pydantic import Field


class Penalty(AI21BaseModel):
    scale: float
    apply_to_whitespaces: Optional[bool] = Field(default=None, alias="applyToWhitespaces")
    apply_to_punctuation: Optional[bool] = Field(default=None, alias="applyToPunctuation")
    apply_to_numbers: Optional[bool] = Field(default=None, alias="applyToNumbers")
    apply_to_stopwords: Optional[bool] = Field(default=None, alias="applyToStopwords")
    apply_to_emojis: Optional[bool] = Field(default=None, alias="applyToEmojis")

    def to_dict(self, **kwargs):
        kwargs["by_alias"] = kwargs.pop("by_alias", True)
        kwargs["exclude_none"] = kwargs.pop("exclude_none", True)

        return super().to_dict(**kwargs)

    def to_json(self, **kwargs) -> str:
        kwargs["by_alias"] = kwargs.pop("by_alias", True)
        kwargs["exclude_none"] = kwargs.pop("exclude_none", True)

        return super().to_json(**kwargs)
